@@ -1,4 +1,4 @@
-# -*- coding: utf-8 -*-
+# -*- coding: future_annotations -*-
 import ast
 from contextlib import contextmanager
 from typing import TYPE_CHECKING
@@ -13,7 +13,7 @@ def _ipython():
     return get_ipython()
 
 
-class _IpythonState(object):
+class _IpythonState:
     def __init__(self):
         self.cell_counter: Optional[int] = None
 
@@ -25,7 +25,7 @@ class _IpythonState(object):
 
     @contextmanager
     def ast_transformer_context(
-            self, transformers: 'Union[List[Union[ast.NodeTransformer, Type]], ast.NodeTransformer, Type]'
+            self, transformers: Union[List[Union[ast.NodeTransformer, Type]], ast.NodeTransformer, Type]
     ):
         if not isinstance(transformers, list):
             transformers = [transformers]
