@@ -31,7 +31,7 @@ from nbsafety.tracing.trace_manager import TraceManager
 # from nbsafety.utils.mixins import EnforceSingletonMixin
 
 if TYPE_CHECKING:
-    from typing import Any, Dict, List, Set, Optional, Tuple, Union
+    from typing import Any, Dict, Iterable, List, Set, Optional, Tuple, Union
     from types import FrameType
     CellId = Union[str, int]
 
@@ -64,7 +64,6 @@ class NotebookSafetySettings(NamedTuple):
     store_history: bool
     test_context: bool
     use_comm: bool
-    trace_messages_enabled: bool
     backwards_cell_staleness_propagation: bool
     track_dependencies: bool
     naive_refresher_computation: bool
@@ -83,7 +82,6 @@ class NotebookSafety(singletons.NotebookSafety):
             store_history=kwargs.pop('store_history', True),
             test_context=kwargs.pop('test_context', False),
             use_comm=use_comm,
-            trace_messages_enabled=kwargs.pop('trace_messages_enabled', False),
             backwards_cell_staleness_propagation=True,
             track_dependencies=True,
             naive_refresher_computation=False,
@@ -91,6 +89,7 @@ class NotebookSafety(singletons.NotebookSafety):
             mode=SafetyRunMode.get(),
         )
         # Note: explicitly adding the types helps PyCharm's built-in code inspection
+        self.trace_messages_enabled: bool = kwargs.pop('trace_messages_enabled', False)
         self.namespaces: Dict[int, NamespaceScope] = {}
         self.aliases: Dict[int, Set[DataSymbol]] = defaultdict(set)
         self.global_scope: Scope = Scope()
@@ -140,6 +139,14 @@ class NotebookSafety(singletons.NotebookSafety):
         else:
             return self._cell_counter
 
+    def reset_cell_counter(self):
+        # only called in test context
+        assert not self.settings.store_history
+        for sym in self.all_data_symbols():
+            sym.last_used_cell_num = sym.defined_cell_num = sym.required_cell_num = 0
+            sym.version_by_used_timestamp.clear()
+        self._cell_counter = 1
+
     def set_ast_transformer_raised(self, new_val: Optional[Exception] = None) -> Optional[Exception]:
         ret = self._ast_transformer_raised
         self._ast_transformer_raised = new_val
@@ -325,9 +332,15 @@ class NotebookSafety(singletons.NotebookSafety):
             'stale': stale_symbols,
         }
 
-    def _precheck_for_stale(self, cell: str):
-        # Precheck process. First obtain the names that need to be checked. Then we check if their
-        # `defined_cell_num` is greater than or equal to required; if not we give a warning and return `True`.
+    def _precheck_for_stale(self, cell: str) -> bool:
+        """
+        This method statically checks the cell to be executed for stale symbols.
+        If any live, stale symbols are detected, it returns `True`. On a syntax
+        error or if no such symbols are detected, return `False`.
+        Furthermore, if this is the second time the user has attempted to execute
+        the exact same code, we assume they want to override this checker and
+        we temporarily mark any stale symbols as being not stale and return `False`.
+        """
         try:
             cell_ast = self._get_cell_ast(cell)
         except SyntaxError:
@@ -338,12 +351,12 @@ class NotebookSafety(singletons.NotebookSafety):
             self._prev_cell_stale_symbols = stale_symbols
             if len(stale_symbols) > 0:
                 warning_counter = 0
-                for node in self._prev_cell_stale_symbols:
+                for sym in self._prev_cell_stale_symbols:
                     if warning_counter >= _MAX_WARNINGS:
                         logger.warning(f'{len(self._prev_cell_stale_symbols) - warning_counter}'
                                        ' more nodes with stale dependencies skipped...')
                         break
-                    _safety_warning(node)
+                    _safety_warning(sym)
                     warning_counter += 1
                 self.stale_dependency_detected = True
                 self._last_refused_code = cell
@@ -351,16 +364,21 @@ class NotebookSafety(singletons.NotebookSafety):
         else:
             # Instead of breaking the dependency chain, simply refresh the nodes
             # with stale deps to their required cell numbers
-            for node in self._prev_cell_stale_symbols:
-                node.defined_cell_num = node.required_cell_num
-                node.namespace_stale_symbols = set()
-                node.fresher_ancestors = set()
+            for sym in self._prev_cell_stale_symbols:
+                sym.temporary_disable_warnings()
             self._prev_cell_stale_symbols.clear()
 
+        # TODO: For each of the live symbols, record their `defined_cell_num` at the
+        #   time of liveness, i.e. at the time `self.cell_counter()`, for use with the
+        #   dynamic slicer.
+        ###########
+        # CODE HERE
+        ###########
+
         self._last_refused_code = None
         return False
 
-    def _resync_symbols(self, symbols: Set[DataSymbol]):
+    def _resync_symbols(self, symbols: Iterable[DataSymbol]):
         for dsym in symbols:
             if not dsym.containing_scope.is_global:
                 continue
@@ -373,9 +391,9 @@ class NotebookSafety(singletons.NotebookSafety):
                 if not alias.containing_scope.is_namespace_scope:
                     continue
                 containing_scope = cast(NamespaceScope, alias.containing_scope)
-                if containing_scope._obj_ref is None or containing_scope._obj_ref() is None:
+                containing_obj = containing_scope.get_obj()
+                if containing_obj is None:
                     continue
-                containing_obj = containing_scope._obj_ref()
                 # TODO: handle dict case too
                 if isinstance(containing_obj, list) and containing_obj[-1] is obj:
                     containing_scope.upsert_data_symbol_for_name(
@@ -395,6 +413,64 @@ class NotebookSafety(singletons.NotebookSafety):
                 del self.namespaces[dsym.obj_id]
                 self.namespaces[namespace.obj_id] = namespace
             dsym.update_obj_ref(obj)
+  
+    def get_cell_dependencies(self, cell_num: int) -> Dict[int, str]:
+        """
+        Gets a dictionary object of cell dependencies for the last or 
+        currently executed cell.
+
+        Args:
+            - cell_num (int): cell to get dependencies for, defaults to last
+                execution counter
+        
+        Returns:
+            - dict (int, str): map from required cell number to code
+                representing dependencies
+        """
+        if cell_num not in self.cell_content_by_counter.keys():
+            raise ValueError(f'Cell {cell_num} has not been run yet.')
+
+        dependencies: Set[int] = set()
+        cell_num_to_dynamic_deps: Dict[int, Set[int]] = defaultdict(set)
+        for sym in self.all_data_symbols():
+            for used_timestamp, version in sym.version_by_used_timestamp.items():
+                cell_num_to_dynamic_deps[used_timestamp].add(version)
+        self._get_cell_dependencies(cell_num, dependencies, cell_num_to_dynamic_deps)
+        return {num: self.cell_content_by_counter[num] for num in dependencies}
+    
+    def _get_cell_dependencies(
+        self, cell_num: int, dependencies: Set[int], cell_num_to_dynamic_deps: Dict[int, Set[int]]
+    ) -> None:
+        """
+        For a given cell, this function recursively populates a set of
+        cell numbers that the given cell depends on, based on the live symbols.
+
+        Args:
+            - dependencies (set<int>): set of cell numbers so far that exist
+            - cell_num (int): current cell to get dependencies for
+            - cell_num_to_ast_used_symbols (dict<int, list<DataSymbol>>): mapping from cell num to symbols last used in said cell
+
+        Returns:
+            None
+        """
+        # Base case: cell already in dependencies
+        if cell_num in dependencies or cell_num <= 0:
+            return
+
+        # Add current cell to dependencies
+        dependencies.add(cell_num)
+
+        # Retrieve cell numbers for the dependent symbols
+        cell = self.cell_content_by_counter[cell_num]
+        live_symbols = self._check_cell_and_resolve_symbols(cell)['live']
+        dep_cell_nums = set(
+            dep_symbol.defined_cell_num for dep_symbol in live_symbols
+        ) | cell_num_to_dynamic_deps[cell_num]
+        logger.info("dynamic cell deps for %d: %s", cell_num, cell_num_to_dynamic_deps[cell_num])
+
+        # For each dependent cell, recursively get their dependencies
+        for num in dep_cell_nums - dependencies:
+            self._get_cell_dependencies(num, dependencies, cell_num_to_dynamic_deps)
 
     def safe_execute(self, cell: str, run_cell_func):
         ret = None
@@ -404,6 +480,7 @@ class NotebookSafety(singletons.NotebookSafety):
             if self._active_cell_id is not None:
                 self._counters_by_cell_id[self._active_cell_id] = self._last_execution_counter
                 self._active_cell_id = None
+
             # Stage 1: Precheck.
             if self._precheck_for_stale(cell) and self.settings.get('skip_unsafe_cells', True):
                 # FIXME: hack to increase cell number
@@ -417,8 +494,11 @@ class NotebookSafety(singletons.NotebookSafety):
                     ret = run_cell_func(cell)
                 # Stage 2.1: resync any defined symbols that could have gotten out-of-sync
                 #  due to tracing being disabled
-                defined = self._check_cell_and_resolve_symbols(cell)['dead']
-                self._resync_symbols(defined)
+
+                self._resync_symbols([
+                    # TODO: avoid bad performance by only iterating over symbols updated in this cell
+                    sym for sym in self.all_data_symbols() if sym.defined_cell_num == self.cell_counter()
+                ])
             finally:
                 if not self.settings.store_history:
                     self._cell_counter += 1
@@ -427,13 +507,12 @@ class NotebookSafety(singletons.NotebookSafety):
     def _make_cell_magic(self, cell_magic_name):
         # this is to avoid capturing `self` and creating an extra reference to the singleton
         store_history = self.settings.store_history
-        self_ = lambda: singletons.nbs()
 
         def _run_cell_func(cell):
             run_cell(cell, store_history=store_history)
 
         def _dependency_safety(_, cell: str):
-            self_().safe_execute(cell, _run_cell_func)
+            singletons.nbs().safe_execute(cell, _run_cell_func)
 
         # FIXME (smacke): probably not a great idea to rely on this
         _dependency_safety.__name__ = cell_magic_name
@@ -447,7 +526,7 @@ class NotebookSafety(singletons.NotebookSafety):
 
         try:
             with TraceManager.instance().tracing_context():
-                with ast_transformer_context([SafetyAstRewriter(self)]):
+                with ast_transformer_context([SafetyAstRewriter()]):
                     yield
         finally:
             # TODO: actually handle errors that occurred in our code while tracing
@@ -465,25 +544,28 @@ class NotebookSafety(singletons.NotebookSafety):
 
         def _safety(line_: str):
             # this is to avoid capturing `self` and creating an extra reference to the singleton
-            self_ = singletons.nbs()
-            line = line_.split()
-            if not line or line[0] not in line_magic_names:
+            cmd, line = line_.split(' ', 1)
+            if cmd in ("show_deps", "show_dependency", "show_dependencies"):
+                return line_magics.show_deps(line)
+            elif cmd == "show_stale":
+                return line_magics.show_stale(line)
+            elif cmd == "trace_messages":
+                return line_magics.trace_messages(line)
+            elif cmd in ("slice", "make_slice", "gather_slice"):
+                return line_magics.make_slice(line)
+            elif cmd == "remove_dependency":
+                return line_magics.remove_dep(line)
+            elif cmd in ("add_dependency", "add_dep"):
+                return line_magics.add_dep(line)
+            elif cmd == "turn_off_warnings_for":
+                return line_magics.turn_off_warnings_for(line)
+            elif cmd == "turn_on_warnings_for":
+                return line_magics.turn_on_warnings_for(line)
+            elif cmd not in line_magic_names:
                 print(line_magics.USAGE)
                 return
-            elif line[0] in ("show_deps", "show_dependency", "show_dependencies"):
-                return line_magics.show_deps(self_, line)
-            elif line[0] == "show_stale":
-                return line_magics.show_stale(self_, line)
-            elif line[0] == "trace_messages":
-                return line_magics.trace_messages(self_, line)
-            elif line[0] == "remove_dependency":
-                return line_magics.remove_dep(self_, line)
-            elif line[0] in ("add_dependency", "add_dep"):
-                return line_magics.add_dep(self_, line)
-            elif line[0] == "turn_off_warnings_for":
-                return line_magics.turn_off_warnings_for(self_, line)
-            elif line[0] == "turn_on_warnings_for":
-                return line_magics.turn_on_warnings_for(self_, line)
+            else:
+                print("We have a magic for %s, but have not yet registered it" % cmd)
 
         # FIXME (smacke): probably not a great idea to rely on this
         _safety.__name__ = _SAFETY_LINE_MAGIC
@@ -503,8 +585,7 @@ class NotebookSafety(singletons.NotebookSafety):
 
     def all_data_symbols(self):
         for alias_set in self.aliases.values():
-            for alias in alias_set:
-                yield alias
+            yield from alias_set
 
     def test_and_clear_detected_flag(self):
         ret = self.stale_dependency_detected
