@@ -3,15 +3,18 @@ import ast
 import logging
 from typing import List, TYPE_CHECKING
 
+from nbsafety.analysis.attr_symbols import resolve_slice_to_constant
 from nbsafety.analysis.mixins import SaveOffAttributesMixin, SkipUnboundArgsMixin, VisitListsMixin
 from nbsafety.data_model.data_symbol import DataSymbol
-from nbsafety.singletons import tracer
+from nbsafety.data_model.scope import NamespaceScope
+from nbsafety.singletons import nbs, tracer
 
 if TYPE_CHECKING:
     from typing import Dict, Optional, Set, Tuple, Union
 
 
 logger = logging.getLogger(__name__)
+logger.setLevel(logging.WARNING)
 
 
 class ResolveRvalSymbols(SaveOffAttributesMixin, SkipUnboundArgsMixin, VisitListsMixin, ast.NodeVisitor):
@@ -63,20 +66,72 @@ class ResolveRvalSymbols(SaveOffAttributesMixin, SkipUnboundArgsMixin, VisitList
         self.visit_AugAssign_or_AnnAssign(node)
 
     def visit_Call(self, node):
-        # TODO: descend further down
+        if isinstance(node.func, (ast.Attribute, ast.Subscript)):
+            self.visit(node.func)
         self.symbols.extend(tracer().resolve_loaded_symbols(node.func))
         self.symbols.extend(tracer().resolve_loaded_symbols(node))
         self.generic_visit([node.args, node.keywords])
 
+    def _get_attr_or_subscript_namespace(self, node: Union[ast.Attribute, ast.Subscript]) -> Optional[NamespaceScope]:
+        with self._push_symbols():
+            self.visit(node.value)
+            symbols = self.symbols
+        if len(symbols) != 1 or symbols[0] is None:
+            return None
+        return nbs().namespaces.get(symbols[0].obj_id, None)
+
     def visit_Attribute(self, node: ast.Attribute):
-        # TODO: we'll ignore args inside of inner calls, e.g. f.g(x, y).h; need to descend further down
-        self.symbols.extend(tracer().resolve_loaded_symbols(node))
+        if isinstance(node.value, ast.Call):
+            self.visit(node.value)
+        symbols = tracer().resolve_loaded_symbols(node)
+        if len(symbols) > 0:
+            self.symbols.extend(symbols)
+            return
+        # TODO: this path lacks coverage
+        try:
+            ns = self._get_attr_or_subscript_namespace(node)
+            if ns is None:
+                return
+            dsym = ns.lookup_data_symbol_by_name_this_indentation(node.attr, is_subscript=False)
+            if dsym is not None:
+                self.symbols.append(dsym)
+        except Exception as e:
+            logger.warning("Exception occurred while resolving node %s: %s", ast.dump(node), e)
 
     def visit_Subscript(self, node: ast.Subscript):
-        # TODO: we'll ignore args inside of inner calls, e.g. f.g(x, y).h; need to descend further down
-        self.symbols.extend(tracer().resolve_loaded_symbols(node))
-        # add slice to RHS to avoid propagating to it
-        self.visit(node.slice)
+        if isinstance(node.value, ast.Call):
+            self.visit(node.value)
+        symbols = tracer().resolve_loaded_symbols(node)
+        with self._push_symbols():
+            # add slice to RHS to avoid propagating to it
+            self.visit(node.slice)
+            symbols.extend(self.symbols)
+        if len(symbols) > 0:
+            self.symbols.extend(symbols)
+            return
+        # TODO: this path lacks coverage
+        try:
+            slice = resolve_slice_to_constant(node)
+            if slice is None or isinstance(slice, ast.Name):
+                return
+            with self._push_symbols():
+                self.visit(node.value)
+                symbols = self.symbols
+            if len(symbols) != 1 or symbols[0] is None:
+                return
+            ns = self._get_attr_or_subscript_namespace(node)
+            if ns is None:
+                return
+            dsym = ns.lookup_data_symbol_by_name_this_indentation(slice, is_subscript=True)
+            if dsym is None and isinstance(slice, int) and slice < 0:
+                try:
+                    dsym = ns.lookup_data_symbol_by_name_this_indentation(len(ns) + slice, is_subscript=True)
+                except TypeError:
+                    dsym = None
+            if dsym is not None:
+                self.symbols.append(dsym)
+        except Exception as e:
+            logger.warning("Exception occurred while resolving node %s: %s", ast.dump(node), e)
 
     def visit_keyword(self, node: ast.keyword):
         self.visit(node.value)
@@ -168,7 +223,21 @@ class ResolveRvalSymbols(SaveOffAttributesMixin, SkipUnboundArgsMixin, VisitList
         pass
 
 
-def resolve_rval_symbols(node: Union[str, ast.AST]) -> Set[DataSymbol]:
+def update_usage_info(symbols: Union[Optional[DataSymbol], Set[Optional[DataSymbol]]]):
+    cell_counter = nbs().cell_counter()
+    for sym in (symbols if symbols is not None and isinstance(symbols, set) else [symbols]):
+        if sym is None:
+            continue
+        sym.last_used_cell_num = cell_counter
+        if sym.defined_cell_num < cell_counter:
+            logger.info('sym `%s` used in cell %d last updated in cell %d', sym, cell_counter, sym.defined_cell_num)
+            sym.version_by_used_timestamp[cell_counter] = sym.defined_cell_num
+
+
+def resolve_rval_symbols(node: Union[str, ast.AST], should_update_usage_info: bool = True) -> Set[DataSymbol]:
     if isinstance(node, str):
         node = ast.parse(node).body[0]
-    return ResolveRvalSymbols()(node)
+    rval_symbols = ResolveRvalSymbols()(node)
+    if should_update_usage_info:
+        update_usage_info(rval_symbols)
+    return rval_symbols
