@@ -1,4 +1,4 @@
-# -*- coding: utf-8 -*-
+# -*- coding: future_annotations -*-
 from enum import Enum
 
 from nbsafety.utils import fast
@@ -13,14 +13,25 @@ class TraceEvent(Enum):
 
     attribute = 'attribute'
     subscript = 'subscript'
-    after_attrsub_chain = 'after_attrsub_chain'
 
-    before_arg_list = 'enter_arg_list'
-    after_arg_list = 'exit_arg_list'
+    before_complex_symbol = 'before_complex_symbol'
+    after_complex_symbol = 'after_complex_symbol'
+
+    before_call = 'before_call'
+    after_call = 'after_call'
     argument = 'argument'
+    before_return = 'before_return'
+    after_return = 'after_return'
 
     before_literal = 'before_literal'
     after_literal = 'after_literal'
+    dict_key = 'dict_key'
+    dict_value = 'dict_value'
+    list_elt = 'list_elt'
+    tuple_elt = 'tuple_elt'
+
+    before_assign_rhs = 'before_assign_rhs'
+    after_assign_rhs = 'after_assign_rhs'
 
     line = 'line'
     call = 'call'
