@@ -1,14 +1,15 @@
-# -*- coding: utf-8 -*-
+# -*- coding: future_annotations -*-
 import logging
 import sys
 
-from test.utils import assert_bool, make_safety_fixture, skipif_known_failing
+from nbsafety.singletons import nbs
+from .utils import assert_bool, make_safety_fixture, skipif_known_failing
 
 logging.basicConfig(level=logging.ERROR)
 
 # Reset dependency graph before each test
-# _safety_fixture, _safety_state, run_cell_ = make_safety_fixture(setup_cells=['%safety trace_messages enable'])
-_safety_fixture, _safety_state, run_cell_ = make_safety_fixture()
+# _safety_fixture, run_cell_ = make_safety_fixture(trace_messages_enabled=True)
+_safety_fixture, run_cell_ = make_safety_fixture()
 
 
 def run_cell(cell, **kwargs):
@@ -21,7 +22,7 @@ def run_cell(cell, **kwargs):
 
 
 def stale_detected():
-    return _safety_state[0].test_and_clear_detected_flag()
+    return nbs().test_and_clear_detected_flag()
 
 
 def assert_detected(msg=''):
@@ -702,9 +703,142 @@ def test_identity_checking_obj():
     assert_not_detected('`y` was not mutated')
 
 
+def test_identity_checking_obj_2():
+    run_cell('x, y = [7], [8]')
+    run_cell('z = y + [3]')
+    run_cell('x[0] = 99')
+    run_cell('logging.info(z)')
+    assert_not_detected('`z` independent of x')
+    run_cell('y[0] = 8')
+    run_cell('logging.info(z)')
+    assert_not_detected('`y` was not mutated')
+    run_cell('y[0] = 42')
+    run_cell('logging.info(z)')
+    assert_detected('`y` was mutated')
+
+
+def test_identity_checking_obj_3():
+    run_cell('d = {"y": 7}')
+    run_cell('x = list(d.values()) + [3]')
+    run_cell('d["y"] = 7')
+    run_cell('logging.info(x)')
+    assert_not_detected('`d["y"]` was not mutated')
+    run_cell('d["y"] = 8')
+    run_cell('logging.info(x)')
+    assert_detected('`d["y"]` was mutated')
+
+
+def test_identity_checking_obj_4():
+    # To get this working properly, we need to create datasyms for literal namespaces recursively
+    run_cell('d = {"y": [7]}')
+    run_cell('d["x"] = d["y"] + [3]')
+    run_cell('d["y"][0] = 7')
+    run_cell('logging.info(d["x"])')
+    assert_not_detected('`d["y"]` was not mutated')
+    run_cell('d["y"][0] = 8')
+    run_cell('logging.info(d["x"])')
+    assert_detected('`d["y"]` was mutated')
+
+
+def test_identity_checking_obj_5():
+    # To get this working properly, we need to create datasyms for literal namespaces recursively
+    run_cell('d = {"y": {0: 7}}')
+    run_cell('d["x"] = {1:3, **d["y"]}')
+    run_cell('d["y"][0] = 7')
+    run_cell('logging.info(d["x"])')
+    assert_not_detected('`d["y"]` was not mutated')
+    run_cell('d["y"][0] = 8')
+    run_cell('logging.info(d["x"])')
+    assert_detected('`d["y"]` was mutated')
+
+
+def test_identity_checking_obj_6():
+    # To get this working properly, we need to create datasyms for literal namespaces recursively
+    run_cell('lst = [[1, 2], 0]')
+    run_cell('lst[1] = lst[0] + [3, 4]')
+    run_cell('lst[0][1] = 2')
+    run_cell('logging.info(lst[1])')
+    assert_not_detected('`lst[0][1]` was not mutated')
+    run_cell('lst[0][1] = 42')
+    run_cell('logging.info(lst[1])')
+    assert_detected('`lst[0][1]` was mutated')
+
+
+def test_starred_assignment_rhs():
+    run_cell('x = 0')
+    run_cell('y = 1')
+    run_cell('z = 2')
+    run_cell('lst = ["foo", "bar"]')
+    # just to make sure the tracer can handle a starred expr in list literal
+    run_cell('a, b, c, d, e = [x + 1, y + 2, z + 3, *lst]')
+    run_cell('z = 42')
+    run_cell('logging.info(a)')
+    assert_not_detected()
+    run_cell('logging.info(b)')
+    assert_not_detected()
+    run_cell('logging.info(c)')
+    assert_detected()
+    run_cell('logging.info(d)')
+    assert_not_detected()
+    run_cell('logging.info(e)')
+    assert_not_detected()
+    run_cell('lst[0] = 7')
+    run_cell('logging.info(d)')
+    assert_detected()
+    run_cell('logging.info(e)')
+    assert_not_detected()
+    run_cell('lst[1] = 8')
+    run_cell('logging.info(e)')
+    assert_detected()
+
+
+def test_starred_assignment():
+    run_cell('x = 0')
+    run_cell('y = 1')
+    run_cell('z = 2')
+    run_cell('lst = ["foo", "bar"]')
+    # just to make sure the tracer can handle a starred expr in list literal
+    run_cell('s, *t = [x + 1, y + 2, z + 3, *lst]')
+    run_cell('z = 42')
+    run_cell('logging.info(s)')
+    assert_not_detected()
+    run_cell('logging.info(t[0])')
+    assert_not_detected()
+    run_cell('logging.info(t[1])')
+    assert_detected()
+    run_cell('x = 99')
+    run_cell('logging.info(s)')
+    assert_detected()
+    run_cell('logging.info(t[0])')
+    assert_not_detected()
+    run_cell('y = 142')
+    run_cell('logging.info(t[0])')
+    assert_detected()
+
+
+def test_starred_assignment_in_middle():
+    run_cell('a, b, c, d, e = 1, 2, 3, 4, 5')
+    run_cell('x, *star, y = [a, b, c, d, e]')
+    run_cell('a += 1')
+    run_cell('logging.info(x)')
+    assert_detected()
+    run_cell('logging.info(star)')
+    assert_not_detected()
+    run_cell('logging.info(y)')
+    assert_not_detected()
+    run_cell('e += 1')
+    run_cell('logging.info(y)')
+    assert_detected()
+    run_cell('logging.info(star)')
+    assert_not_detected()
+    run_cell('c += 1')
+    run_cell('logging.info(star)')
+    assert_detected()
+
+
 def test_attributes():
     run_cell("""
-class Foo(object):
+class Foo:
     def __init__(self, x):
         self.x = x
 """)
@@ -718,7 +852,7 @@ class Foo(object):
 @skipif_known_failing
 def test_attributes_2():
     run_cell("""
-class Foo(object):
+class Foo:
     def __init__(self, x):
         self.x = x
 """)
@@ -731,9 +865,66 @@ class Foo(object):
     assert_detected('y depends on stale x')
 
 
+def test_attribute_unpacking():
+    run_cell("""
+class Foo:
+    def __init__(self, x):
+        self.x = x
+""")
+    run_cell('x = Foo(5)')
+    run_cell('y = Foo(6)')
+    run_cell('w = 42')
+    run_cell('z = 43')
+    run_cell('x.x, y.x = w + 2, z + 3')
+    run_cell('z = 9001')
+    run_cell('logging.info(x.x)')
+    assert_not_detected()
+    run_cell('logging.info(x)')
+    assert_not_detected()
+    run_cell('logging.info(y.x)')
+    assert_detected()
+    run_cell('logging.info(y)')
+    assert_detected()
+    run_cell('y.x = z + 3')
+    run_cell('logging.info(y.x)')
+    assert_not_detected()
+    run_cell('w = 99')
+    run_cell('logging.info(x.x)')
+    assert_detected()
+    run_cell('logging.info(x)')
+    assert_detected()
+    run_cell('logging.info(y.x)')
+    assert_not_detected()
+    run_cell('logging.info(y)')
+    assert_not_detected()
+
+
+def test_attribute_unpacking_no_overwrite():
+    run_cell("""
+class Foo:
+    def __init__(self, x):
+        self.x = x
+""")
+    run_cell('x = Foo(5)')
+    run_cell('y = Foo(6)')
+    run_cell('w = 42')
+    run_cell('z = 43')
+    run_cell('x.x, y.x = w + 2, z + 3')
+    run_cell('s, t = 12, 13')
+    run_cell('x.x, y.x = x.x + s, y.x + t')
+    run_cell('w = 101')
+    run_cell('logging.info(x.x)')
+    assert_detected()
+    run_cell('logging.info(y.x)')
+    assert_not_detected()
+    run_cell('z = 103')
+    run_cell('logging.info(y.x)')
+    assert_detected()
+
+
 def test_attributes_3():
     run_cell("""
-class Foo(object):
+class Foo:
     def __init__(self, x, y):
         self.x = x
         self.y = y
@@ -748,7 +939,7 @@ class Foo(object):
 
 def test_stale_use_of_attribute():
     run_cell("""
-class Foo(object):
+class Foo:
     def __init__(self, x, y):
         self.x = x
         self.y = y
@@ -764,7 +955,7 @@ class Foo(object):
 def test_attr_manager_active_scope_resets():
     run_cell("""
 y = 10
-class Foo(object):
+class Foo:
     def f(self):
         y = 11
         return y
@@ -783,7 +974,7 @@ def f():
 def test_attribute_call_point():
     run_cell("""
 y = 10
-class Foo(object):
+class Foo:
     def f(self):
         return y
 """)
@@ -797,7 +988,7 @@ class Foo(object):
 def test_attr_manager_active_scope_with_property():
     run_cell("""
 y = 10
-class Foo(object):
+class Foo:
     @property
     def f(self):
         y = 11
@@ -815,7 +1006,7 @@ class Foo(object):
 def test_namespace_scope_resolution():
     run_cell("""
 y = 42
-class Foo(object):
+class Foo:
     y = 10
     @property
     def foo(self):
@@ -830,13 +1021,13 @@ class Foo(object):
 
 def test_long_chain_attribute():
     run_cell("""
-class Foo(object):
+class Foo:
     shared = 99
     def __init__(self, x, y):
         self.x = x
         self.y = y + self.shared
         
-    class Bar(object):
+    class Bar:
         def __init__(self, a, b):
             self.a = a
             self.b = b
@@ -949,7 +1140,7 @@ def test_subscript_sensitivity():
     run_cell('lst[i] = 10')
     run_cell('i = 1')
     run_cell('logging.info(lst)')
-    assert_detected('lst depends on stale i')
+    assert_detected('`lst[0]` depends on stale i')
 
 
 def test_subscript_adds():
@@ -998,7 +1189,7 @@ for word in s.split('X'):
 
 def test_lazy_class_scope_resolution():
     run_cell("""
-class Foo(object):
+class Foo:
     shared = 99
     def __init__(self, x):
         self.x = x
@@ -1017,7 +1208,7 @@ class Foo(object):
 
 def test_new_scope_val_depends_on_old():
     run_cell("""
-class Foo(object):
+class Foo:
     shared = 99
 """)
     run_cell('foo = Foo()')
@@ -1034,7 +1225,7 @@ class Foo(object):
 
 def test_class_member_mutation_does_not_affect_instance_members():
     run_cell("""
-class Foo(object):
+class Foo:
     shared = 99
     def __init__(self):
         self.x = 42
@@ -1171,10 +1362,11 @@ def test_single_line_dictionary_literal_fix_stale_deps():
     run_cell('logging.info(d)')
     assert_detected('`d` depends on stale `bar`')
     run_cell('d[foo] = bar')
+    assert_not_detected()
     run_cell('logging.info(d)')
-    assert_false_positive('`d`s stale dep fixed, but this is hard to detect '
-                          'since we did not yet have a DataSymbol for `d[foo]` when staleness introduced')
-    # assert_not_detected('`d`s stale dep fixed')
+    # assert_false_positive('`d`s stale dep fixed, but this is hard to detect '
+    #                       'since we did not yet have a DataSymbol for `d[foo]` when staleness introduced')
+    assert_not_detected('`d`s stale dep fixed')
     run_cell('foo = 8')
     run_cell('logging.info(d)')
     assert_detected('`d` depends on stale `foo`')
@@ -1318,23 +1510,15 @@ logging.info(x)
 
 
 def test_exception_stack_unwind():
-    import builtins
-    safety_state = '_safety_state'
-    test_passed = 'test_passed'
-    setattr(builtins, safety_state, _safety_state[0])
-    setattr(builtins, test_passed, True)
-    test_passed = "'" + test_passed + "'"
-
     def assert_stack_size(size):
-        return ';'.join([
-            f'can_pass = len({safety_state}.tracing_manager._stack) == {size}',
-            f'setattr(builtins, {test_passed}, getattr(builtins, {test_passed}) and can_pass)',
-            f'print(len({safety_state}.tracing_manager._stack), "vs", {size})'
+        len_call_stack = 'len(TraceManager.instance().call_stack)'
+        return ', '.join([
+            f'assert {len_call_stack} == {size}',
+            f'"%d vs {size}" % {len_call_stack}',
         ])
-    try:
-        run_cell(f"""
-import builtins
+    run_cell(f"""
 import numpy as np
+from nbsafety.singletons import TraceManager
 {assert_stack_size(0)}
 def f():
     {assert_stack_size(1)}
@@ -1351,13 +1535,6 @@ def f():
 f()
 {assert_stack_size(0)}
 """)
-    finally:
-        delattr(builtins, safety_state)
-    test_passed = test_passed.strip("'")
-    try:
-        assert getattr(builtins, test_passed), 'unexpected stack size somewhere'
-    finally:
-        delattr(builtins, test_passed)
 
 
 def test_throwing_statements_do_not_track_deps():
@@ -1436,6 +1613,17 @@ except:
     assert_not_detected('x inside class scope is different')
 
 
+def test_set_literal():
+    run_cell('x, y, z = 1, 2, 3')
+    run_cell('s = {x + 1, y + 7}')
+    run_cell('z = 42')
+    run_cell('logging.info(s)')
+    assert_not_detected()
+    run_cell('x = 17')
+    run_cell('logging.info(s)')
+    assert_detected()
+
+
 def test_tuple_unpack_simple():
     run_cell('x, y = 0, 1')
     run_cell('a, b = x + 2, y + 3')
@@ -1462,6 +1650,17 @@ def test_tuple_unpack_hard():
     assert_detected('`a` depends on stale `x`')
 
 
+def test_unpack_multiple_from_single():
+    run_cell('x, y = 0, 1')
+    run_cell('lst = [x + 1, y + 1]')
+    run_cell('a, b = lst')
+    run_cell('x = 42')
+    run_cell('logging.info(b)')
+    assert_not_detected()
+    run_cell('logging.info(a)')
+    assert_detected()
+
+
 @skipif_known_failing
 def test_attr_dep_with_top_level_overwrite():
     run_cell("""
@@ -1620,11 +1819,11 @@ def f(p):
     assert_not_detected()
     run_cell('x = 3')
     run_cell('logging.info(z)')
-    assert_false_negative('tracing should have been reactivated but this is hard')
+    assert_detected('tracing should not be disabled')
 
 
 def test_tracing_disable_with_nested_calls():
-    run_cell('%safety trace_messages enable')
+    # run_cell('%safety trace_messages enable')
     run_cell('y = 0')
     run_cell("""
 def f():
@@ -1643,10 +1842,19 @@ x = g(True) + 1
 """)
     run_cell('y = 42')
     run_cell('logging.info(x)')
-    assert_false_negative('`x` has dep on stale `y` but capturing this is hard')
+    assert_detected('`x` has dep on stale `y`')
 
 
 def test_dict():
+    run_cell('d = {}; d[0] = 0')
+    run_cell('x = d[0] + 1')
+    run_cell('d[0] = 42')
+    run_cell('logging.info(x)')
+    assert_detected('`x` has dependency on old value of `d[0]`')
+
+
+@skipif_known_failing
+def test_dict_2():
     run_cell('d = {}; d[0] = 0')
     run_cell('x = d[0] + 1')
     run_cell('d = {}; d[0] = 42')
@@ -1660,7 +1868,7 @@ def test_default_dict():
     run_cell('x = d[0][0] + 1')
     run_cell('d = defaultdict(dict); d[0][0] = 42')
     run_cell('logging.info(x)')
-    assert_detected('`x` has dependency on old value of `d[0][0]`')
+    assert_false_negative('`x` has dependency on old value of `d[0][0]`, but this is hard to detect b/c d was cleared')
 
 
 def test_mutate_arg():
@@ -1688,7 +1896,7 @@ def test_mutate_arg_special_cases():
     assert_not_detected('`logging.info` does not mutate `y`')
 
 
-def test_augassign_does_not_overwrite():
+def test_augassign_does_not_overwrite_deps():
     run_cell('x = 0')
     run_cell('y = 1')
     run_cell('z = x + 2')
@@ -1698,6 +1906,44 @@ def test_augassign_does_not_overwrite():
     assert_detected('`z` depends on old value of `x`')
 
 
+def test_rval_appearance_does_not_overwrite_deps():
+    run_cell('x = 0')
+    run_cell('y = 1')
+    run_cell('z = x + 2')
+    run_cell('z = z + y')
+    run_cell('x = 42')
+    run_cell('logging.info(z)')
+    assert_detected('`z` depends on old value of `x`')
+
+
+def test_augassign_does_not_overwrite_deps_for_attributes():
+    run_cell("""
+class Foo:
+    def __init__(self, x):
+        self.x = x
+""")
+    run_cell('x, y, z = Foo(0), Foo(5), Foo(10)')
+    run_cell('z.x = x.x + 2')
+    run_cell('z.x += y.x')
+    run_cell('x.x = 42')
+    run_cell('logging.info(z.x)')
+    assert_detected('`z.x` depends on old value of `x.x`')
+
+
+def test_rval_appearance_does_not_overwrite_deps_for_attributes():
+    run_cell("""
+class Foo:
+    def __init__(self, x):
+        self.x = x
+""")
+    run_cell('x, y, z = Foo(0), Foo(5), Foo(10)')
+    run_cell('z.x = x.x + 2')
+    run_cell('z.x = z.x + y.x')
+    run_cell('x.x = 42')
+    run_cell('logging.info(z.x)')
+    assert_detected('`z.x` depends on old value of `x.x`')
+
+
 def test_context_manager():
     run_cell("""
 from contextlib import contextmanager
@@ -1736,6 +1982,105 @@ y = x + 1
     assert_detected('`y` depends on old value `x`')
 
 
+def test_tuple_return():
+    run_cell("""
+x = 0
+y = 1
+a = x + 42
+b = y + 77
+def foo():
+    return a, b
+""")
+    run_cell('t = foo()[1]')
+    run_cell('x = 9')
+    run_cell('logging.info(t)')
+    assert_not_detected('`t` independent of updated `x`')
+    run_cell('y = 10')
+    run_cell('logging.info(t)')
+    assert_detected('`t` depends on old version of `y`')
+
+
+def test_tuple_return_obj():
+    run_cell("""
+x = 0
+y = 1
+a = x + 42
+b = y + 77
+def foo():
+    return [a], [b]
+""")
+    run_cell('t = foo()[1][0]')
+    run_cell('x = 9')
+    run_cell('logging.info(t)')
+    assert_not_detected('`t` independent of updated `x`')
+    run_cell('y = 10')
+    run_cell('logging.info(t)')
+    assert_detected('`t` depends on old version of `y`')
+
+
+def test_property_in_function_arg():
+    run_cell("""
+z = 42
+class Foo:
+    @property
+    def bar(self):
+        return z
+
+def f(x, y):
+    return x + y
+""")
+    run_cell('foo = Foo()')
+    run_cell('w = f(3, foo.bar)')
+    run_cell('z = 7')
+    run_cell('logging.info(w)')
+    assert_detected('`w` depends on old version of `z`')
+
+
+def test_getter_setter_with_global():
+    run_cell("""
+z = 42
+class Bar:
+    shared = 25
+    @property
+    def baz(self):
+        return z
+    
+    @baz.setter
+    def baz(self, new_val):
+        global z
+        z = new_val
+    
+class Foo:
+    def __init__(self):
+        self.bar = Bar()
+
+def f(x, y):
+    return x + y
+""")
+    run_cell('foo = Bar()')
+    run_cell('w = f(3, foo.baz)')
+    run_cell('logging.info(w)')
+    assert_not_detected()
+    run_cell('foo.baz = 84')
+    run_cell('logging.info(w)')
+    assert_detected('`w` depends on stale `foo.baz`')
+
+
+@skipif_known_failing
+def test_list_sum_simple():
+    run_cell("w, x, y, z = 0, 1, 2, 3")
+    run_cell("lst = [w, x] + [y, z]")
+    run_cell("z += 42")
+    run_cell("logging.info(lst[0])")
+    assert_not_detected()
+    run_cell("logging.info(lst[1])")
+    assert_not_detected()
+    run_cell("logging.info(lst[2])")
+    assert_not_detected()
+    run_cell("logging.info(lst[3])")
+    assert_detected()
+
+
 if sys.version_info >= (3, 8):
     def test_walrus_simple():
         run_cell("""
@@ -1760,3 +2105,24 @@ if (x := (y := (z := 1) + 1) + 1) > 0:
         assert_detected('`x` depends on old value of `y`')
         run_cell('logging.info(a)')
         assert_detected('`a` depends on old value of `y`')
+
+    def test_walrus_fancy_attributes():
+        run_cell("""
+class Foo:
+    def __init__(self, x):
+        self.x = x
+""")
+        run_cell('foo = Foo(9001)')
+        run_cell("""
+if (x := (y := (z := 1) + foo.x) + 1) > 0:
+    a = x + 1
+""")
+        run_cell('foo.x = 42')
+        run_cell('logging.info(z)')
+        assert_not_detected('`z` does not depend on `foo.x`')
+        run_cell('logging.info(y)')
+        assert_detected('`y` depends on old `foo.x`')
+        run_cell('logging.info(x)')
+        assert_detected('`x` depends on old value of `foo.x`')
+        run_cell('logging.info(a)')
+        assert_detected('`a` depends on old value of `foo.x`')
